@@ -4,10 +4,17 @@ extern crate syn;
 #[macro_use]
 extern crate quote;
 
+use concordium_contracts_common::*;
 use proc_macro::TokenStream;
 use proc_macro2::Span;
 use quote::ToTokens;
-use std::{convert::TryFrom, ops::Neg};
+#[cfg(feature = "build-schema")]
+use std::collections::HashMap;
+use std::{
+    collections::{BTreeMap, BTreeSet},
+    convert::TryFrom,
+    ops::Neg,
+};
 use syn::{
     parse::Parser, parse_macro_input, punctuated::*, spanned::Spanned, DataEnum, Ident, Meta, Token,
 };
@@ -30,47 +37,298 @@ fn attach_error<A>(mut v: syn::Result<A>, msg: &str) -> syn::Result<A> {
     v
 }
 
-/// Get the name item from a list, if available and a string literal.
-/// If the named item does not have the expected (string) value, this will
-/// return an Err. If the item does not exist the return value is Ok(None).
-/// FIXME: Ensure there is only one.
-fn get_attribute_value<'a, I: IntoIterator<Item = &'a Meta>>(
-    iter: I,
-    name: &str,
-) -> syn::Result<Option<String>> {
+/// Attributes that can be attached either to the init or receive method of a
+/// smart contract.
+struct OptionalArguments {
+    /// If set, the contract can receive gTU.
+    pub(crate) payable:       bool,
+    /// If enabled, the function has access to logging facilities.
+    pub(crate) enable_logger: bool,
+    /// The function is a low-level one, with direct access to contract memory.
+    pub(crate) low_level:     bool,
+    /// Which type, if any, is the parameter type of the contract.
+    /// This is used when generating schemas.
+    pub(crate) parameter:     Option<syn::LitStr>,
+}
+
+/// Attributes that can be attached to the initialization method.
+struct InitAttributes {
+    /// Name of the contract.
+    pub(crate) contract: syn::LitStr,
+    pub(crate) optional: OptionalArguments,
+}
+
+/// Attributes that can be attached to the receive method.
+struct ReceiveAttributes {
+    /// Name of the contract the method applies to.
+    pub(crate) contract: syn::LitStr,
+    /// Name of the method.
+    pub(crate) name:     syn::LitStr,
+    pub(crate) optional: OptionalArguments,
+}
+
+#[derive(Default)]
+struct ParsedAttributes {
+    /// We use BTreeSet to have consistent order of iteration when reporting
+    /// errors.
+    pub(crate) flags:  BTreeSet<syn::Ident>,
+    /// We use BTreeMap to have consistent order of iteration when reporting
+    /// errors.
+    pub(crate) values: BTreeMap<syn::Ident, syn::LitStr>,
+}
+
+impl ParsedAttributes {
+    /// Remove an attribute and return it, if present. The key must be a valid
+    /// Rust identifier, otherwise this function will panic.
+    pub(crate) fn extract_value(&mut self, key: &str) -> Option<syn::LitStr> {
+        // This is not clean, constructing a new identifier with a call_site span.
+        // But the only alternative I see is iterating over the map and locating the key
+        // since Ident implements equality comparison with &str.
+        let key = syn::Ident::new(key, Span::call_site());
+        self.values.remove(&key)
+    }
+
+    /// Remove an attribute and return whether it was present.
+    pub(crate) fn extract_flag(&mut self, key: &str) -> bool {
+        // This is not clean, constructing a new identifier with a call_site span.
+        // But the only alternative I see is iterating over the map and locating the key
+        // since Ident implements equality comparison with &str.
+        let key = syn::Ident::new(key, Span::call_site());
+        self.flags.remove(&key)
+    }
+
+    /// If there are any remaining attributes signal an error. Otherwise return
+    /// Ok(())
+    pub(crate) fn report_all_attributes(self) -> syn::Result<()> {
+        // TODO: Replace into_iter + map with into_keys when only supporting rust 1.54+
+        let mut iter = self.flags.into_iter().chain(self.values.into_iter().map(|(k, _)| k));
+        if let Some(ident) = iter.next() {
+            let mut err =
+                syn::Error::new(ident.span(), format!("Unrecognized attribute {}.", ident));
+            for next_ident in iter {
+                err.combine(syn::Error::new(
+                    ident.span(),
+                    format!("Unrecognized attribute {}.", next_ident),
+                ));
+            }
+            Err(err)
+        } else {
+            Ok(())
+        }
+    }
+}
+
+/// Parse attributes ensuring there are no duplicate items.
+fn parse_attributes<'a>(iter: impl IntoIterator<Item = &'a Meta>) -> syn::Result<ParsedAttributes> {
+    let mut ret = ParsedAttributes::default();
+    let mut errors = Vec::new();
+    let mut duplicate_values = BTreeMap::new();
+    let mut duplicate_flags = BTreeMap::new();
     for attr in iter.into_iter() {
         match attr {
             Meta::NameValue(mnv) => {
-                if mnv.path.is_ident(name) {
-                    if let syn::Lit::Str(lit) = &mnv.lit {
-                        return Ok(Some(lit.value()));
+                if let Some(ident) = mnv.path.get_ident() {
+                    if let syn::Lit::Str(ls) = &mnv.lit {
+                        if let Some((existing_ident, _)) = ret.values.get_key_value(ident) {
+                            let v = duplicate_values.entry(ident).or_insert_with(|| {
+                                syn::Error::new(
+                                    existing_ident.span(),
+                                    format!("Duplicate attribute '{}'.", existing_ident),
+                                )
+                            });
+                            v.combine(syn::Error::new(
+                                ident.span(),
+                                format!("'{}' also appears here.", ident),
+                            ));
+                        } else {
+                            ret.values.insert(ident.clone(), ls.clone());
+                        }
                     } else {
-                        return Err(syn::Error::new(
-                            mnv.span(),
-                            format!("The `{}` attribute must be a string literal.", name),
+                        errors.push(syn::Error::new(
+                            mnv.path.span(),
+                            format!(
+                                "Values of attribute must be string literals, e.g., '{} = \
+                                 \"value\"'",
+                                ident
+                            ),
                         ));
                     }
+                } else {
+                    errors.push(syn::Error::new(
+                        mnv.path.span(),
+                        "Unrecognized attribute. Only attribute names consisting of a single \
+                         identifier are recognized.",
+                    ))
                 }
             }
             Meta::Path(p) => {
-                if p.is_ident(name) {
-                    return Err(syn::Error::new(
-                        attr.span(),
-                        format!("The `{}` attribute must have a string literal value.", name),
-                    ));
+                if let Some(ident) = p.get_ident() {
+                    if let Some(existing_ident) = ret.flags.get(ident) {
+                        let v = duplicate_flags.entry(ident).or_insert_with(|| {
+                            syn::Error::new(
+                                existing_ident.span(),
+                                format!("Duplicate attribute '{}'.", existing_ident),
+                            )
+                        });
+                        v.combine(syn::Error::new(
+                            ident.span(),
+                            format!("'{}' also appears here.", ident),
+                        ));
+                    } else {
+                        ret.flags.insert(ident.clone());
+                    }
+                } else {
+                    errors.push(syn::Error::new(
+                        p.span(),
+                        "Unrecognized attribute. Only attribute names consisting of a single \
+                         identifier are recognized.",
+                    ))
                 }
             }
             Meta::List(p) => {
-                if p.path.is_ident(name) {
-                    return Err(syn::Error::new(
-                        attr.span(),
-                        format!("The `{}` attribute must have a string literal value.", name),
-                    ));
-                }
+                errors.push(syn::Error::new(p.span(), "Unrecognized attribute."));
             }
         }
     }
-    Ok(None)
+    // TODO: Replace with into_values when least rust version becomes 1.54.
+    let mut iter = errors
+        .into_iter()
+        .chain(duplicate_values.into_iter().map(|(_, v)| v))
+        .chain(duplicate_flags.into_iter().map(|(_, v)| v));
+    // If there are any errors we combine them.
+    if let Some(err) = iter.next() {
+        let mut err = err;
+        for next_err in iter {
+            err.combine(next_err);
+        }
+        Err(err)
+    } else {
+        Ok(ret)
+    }
+}
+
+#[cfg(feature = "build-schema")]
+/// Attributes applicable to the `contract_state` annotation.
+struct ContractStateAttributes {
+    /// Name of the contract the contract state applies to.
+    pub(crate) contract: syn::LitStr,
+}
+
+#[cfg(feature = "build-schema")]
+// Attribute names for the contract_state macro.
+const CONTRACT_STATE_ATTRIBUTE_CONTRACT: &str = "contract";
+
+#[cfg(feature = "build-schema")]
+/// Parse nested attributes to the `contract_state` attribute.
+fn parse_contract_state_attributes<'a, I: IntoIterator<Item = &'a Meta>>(
+    attrs: I,
+) -> syn::Result<ContractStateAttributes> {
+    let mut attributes = parse_attributes(attrs)?;
+    let contract =
+        attributes.extract_value(CONTRACT_STATE_ATTRIBUTE_CONTRACT).ok_or_else(|| {
+            syn::Error::new(
+                Span::call_site(),
+                "A name for the contract must be provided, using the 'contract' attribute.\n\nFor \
+                 example, #[contract_state(contract = \"my-contract\")]",
+            )
+        })?;
+    attributes.report_all_attributes()?;
+    Ok(ContractStateAttributes {
+        contract,
+    })
+}
+
+// Supported attributes for the init methods.
+
+const INIT_ATTRIBUTE_PARAMETER: &str = "parameter";
+const INIT_ATTRIBUTE_CONTRACT: &str = "contract";
+const INIT_ATTRIBUTE_PAYABLE: &str = "payable";
+const INIT_ATTRIBUTE_ENABLE_LOGGER: &str = "enable_logger";
+const INIT_ATTRIBUTE_LOW_LEVEL: &str = "low_level";
+
+fn parse_init_attributes<'a, I: IntoIterator<Item = &'a Meta>>(
+    attrs: I,
+) -> syn::Result<InitAttributes> {
+    let mut attributes = parse_attributes(attrs)?;
+    let contract: syn::LitStr =
+        attributes.extract_value(INIT_ATTRIBUTE_CONTRACT).ok_or_else(|| {
+            syn::Error::new(
+                Span::call_site(),
+                "A name for the contract must be provided, using the 'contract' attribute.\n\nFor \
+                 example, #[init(contract = \"my-contract\")]",
+            )
+        })?;
+    let parameter: Option<syn::LitStr> = attributes.extract_value(INIT_ATTRIBUTE_PARAMETER);
+    let payable = attributes.extract_flag(INIT_ATTRIBUTE_PAYABLE);
+    let enable_logger = attributes.extract_flag(INIT_ATTRIBUTE_ENABLE_LOGGER);
+    let low_level = attributes.extract_flag(INIT_ATTRIBUTE_LOW_LEVEL);
+    // Make sure that there are no unrecognized attributes. These would typically be
+    // there due to an error. An improvement would be to find the nearest valid one
+    // for each of them and report that in the error.
+    attributes.report_all_attributes()?;
+    Ok(InitAttributes {
+        contract,
+        optional: OptionalArguments {
+            payable,
+            enable_logger,
+            low_level,
+            parameter,
+        },
+    })
+}
+
+// Supported attributes for the receive methods.
+
+const RECEIVE_ATTRIBUTE_PARAMETER: &str = "parameter";
+const RECEIVE_ATTRIBUTE_CONTRACT: &str = "contract";
+const RECEIVE_ATTRIBUTE_NAME: &str = "name";
+const RECEIVE_ATTRIBUTE_PAYABLE: &str = "payable";
+const RECEIVE_ATTRIBUTE_ENABLE_LOGGER: &str = "enable_logger";
+const RECEIVE_ATTRIBUTE_LOW_LEVEL: &str = "low_level";
+
+fn parse_receive_attributes<'a, I: IntoIterator<Item = &'a Meta>>(
+    attrs: I,
+) -> syn::Result<ReceiveAttributes> {
+    let mut attributes = parse_attributes(attrs)?;
+
+    let contract = attributes.extract_value(RECEIVE_ATTRIBUTE_CONTRACT);
+    let name = attributes.extract_value(RECEIVE_ATTRIBUTE_NAME);
+    let parameter: Option<syn::LitStr> = attributes.extract_value(RECEIVE_ATTRIBUTE_PARAMETER);
+    let payable = attributes.extract_flag(RECEIVE_ATTRIBUTE_PAYABLE);
+    let enable_logger = attributes.extract_flag(RECEIVE_ATTRIBUTE_ENABLE_LOGGER);
+    let low_level = attributes.extract_flag(RECEIVE_ATTRIBUTE_LOW_LEVEL);
+    // Make sure that there are no unrecognized attributes. These would typically be
+    // there due to an error. An improvement would be to find the nearest valid one
+    // for each of them and report that in the error.
+    attributes.report_all_attributes()?;
+    match (contract, name) {
+        (Some(contract), Some(name)) => Ok(ReceiveAttributes {
+            contract,
+            name,
+            optional: OptionalArguments {
+                payable,
+                enable_logger,
+                low_level,
+                parameter,
+            },
+        }),
+        (Some(_), None) => Err(syn::Error::new(
+            Span::call_site(),
+            "A name for the method must be provided, using the 'name' attribute.\n\nFor example, \
+             #[receive(name = \"receive\")]",
+        )),
+        (None, Some(_)) => Err(syn::Error::new(
+            Span::call_site(),
+            "A name for the method must be provided, using the 'contract' attribute.\n\nFor \
+             example, #[receive(contract = \"my-contract\")]",
+        )),
+        (None, None) => Err(syn::Error::new(
+            Span::call_site(),
+            "A contract name and a name for the method must be provided, using the 'contract' and \
+             'name' attributes.\n\nFor example, #[receive(contract = \"my-contract\", name = \
+             \"receive\")]",
+        )),
+    }
 }
 
 // Return whether a attribute item is present.
@@ -164,22 +422,23 @@ pub fn init(attr: TokenStream, item: TokenStream) -> TokenStream {
 }
 
 fn init_worker(attr: TokenStream, item: TokenStream) -> syn::Result<TokenStream> {
-    let attrs = Punctuated::<Meta, Token![,]>::parse_terminated.parse(attr)?;
-
-    let contract_name = get_attribute_value(attrs.iter(), "contract")?.ok_or_else(|| {
-        syn::Error::new(
-            attrs.span(),
-            "A name for the contract must be provided, using the contract attribute. For example, \
-             #[init(contract = \"my-contract\")]",
-        )
-    })?;
-
     let ast: syn::ItemFn =
         attach_error(syn::parse(item), "#[init] can only be applied to functions.")?;
 
+    let attrs = Punctuated::<Meta, Token![,]>::parse_terminated.parse(attr)?;
+
+    let init_attributes = parse_init_attributes(&attrs)?;
+
+    let contract_name = init_attributes.contract;
+
     let fn_name = &ast.sig.ident;
     let rust_export_fn_name = format_ident!("export_{}", fn_name);
-    let wasm_export_fn_name = format!("init_{}", contract_name);
+    let wasm_export_fn_name = format!("init_{}", contract_name.value());
+
+    if let Err(e) = ContractName::is_valid_contract_name(&wasm_export_fn_name) {
+        return Err(syn::Error::new(contract_name.span(), e));
+    }
+
     let amount_ident = format_ident!("amount");
 
     // Accumulate a list of required arguments, if the function contains a
@@ -187,10 +446,13 @@ fn init_worker(attr: TokenStream, item: TokenStream) -> syn::Result<TokenStream>
     // strings are displayed as the expected arguments.
     let mut required_args = vec!["ctx: &impl HasInitContext"];
 
-    let (setup_fn_optional_args, fn_optional_args) =
-        contract_function_optional_args_tokens(&attrs, &amount_ident, &mut required_args);
+    let (setup_fn_optional_args, fn_optional_args) = contract_function_optional_args_tokens(
+        &init_attributes.optional,
+        &amount_ident,
+        &mut required_args,
+    );
 
-    let mut out = if contains_attribute(attrs.iter(), "low_level") {
+    let mut out = if init_attributes.optional.low_level {
         required_args.push("state: &mut ContractState");
         quote! {
             #[export_name = #wasm_export_fn_name]
@@ -252,7 +514,7 @@ fn init_worker(attr: TokenStream, item: TokenStream) -> syn::Result<TokenStream>
     }
 
     // Embed schema if 'parameter' attribute is set
-    let parameter_option = get_attribute_value(attrs.iter(), "parameter")?;
+    let parameter_option = init_attributes.optional.parameter.map(|a| a.value());
     out.extend(contract_function_schema_tokens(
         parameter_option,
         rust_export_fn_name,
@@ -353,30 +615,40 @@ pub fn receive(attr: TokenStream, item: TokenStream) -> TokenStream {
 }
 
 fn receive_worker(attr: TokenStream, item: TokenStream) -> syn::Result<TokenStream> {
-    let attrs = Punctuated::<Meta, Token![,]>::parse_terminated.parse(attr)?;
-
-    let contract_name = get_attribute_value(attrs.iter(), "contract")?.ok_or_else(|| {
-        syn::Error::new(
-            attrs.span(),
-            "The name of the associated contract must be provided, using the 'contract' \
-             attribute.\n\nFor example, #[receive(contract = \"my-contract\")]",
-        )
-    })?;
-
-    let name = get_attribute_value(attrs.iter(), "name")?.ok_or_else(|| {
-        syn::Error::new(
-            attrs.span(),
-            "A name for the receive function must be provided, using the 'name' attribute.\n\nFor \
-             example, #[receive(name = \"func-name\", ...)]",
-        )
-    })?;
-
     let ast: syn::ItemFn =
         attach_error(syn::parse(item), "#[receive] can only be applied to functions.")?;
 
+    let attrs = Punctuated::<Meta, Token![,]>::parse_terminated.parse(attr)?;
+
+    let receive_attributes = parse_receive_attributes(&attrs)?;
+
+    let contract_name = receive_attributes.contract;
+
+    let method_name = receive_attributes.name;
+
     let fn_name = &ast.sig.ident;
     let rust_export_fn_name = format_ident!("export_{}", fn_name);
-    let wasm_export_fn_name = format!("{}.{}", contract_name, name);
+    let wasm_export_fn_name = format!("{}.{}", contract_name.value(), method_name.value());
+
+    // Validate the contract name independently to ensure that it doesn't contain a
+    // '.' as this causes a subtle error when receive names are being split.
+    let contract_name_validation =
+        ContractName::is_valid_contract_name(&format!("init_{}", contract_name.value()))
+            .map_err(|e| syn::Error::new(contract_name.span(), e));
+
+    let receive_name_validation = ReceiveName::is_valid_receive_name(&wasm_export_fn_name)
+        .map_err(|e| syn::Error::new(method_name.span(), e));
+
+    match (contract_name_validation, receive_name_validation) {
+        (Err(mut e0), Err(e1)) => {
+            e0.combine(e1);
+            return Err(e0);
+        }
+        (Err(e), _) => return Err(e),
+        (_, Err(e)) => return Err(e),
+        _ => (),
+    };
+
     let amount_ident = format_ident!("amount");
 
     // Accumulate a list of required arguments, if the function contains a
@@ -384,10 +656,13 @@ fn receive_worker(attr: TokenStream, item: TokenStream) -> syn::Result<TokenStre
     // strings are displayed as the expected arguments.
     let mut required_args = vec!["ctx: &impl HasReceiveContext"];
 
-    let (setup_fn_optional_args, fn_optional_args) =
-        contract_function_optional_args_tokens(&attrs, &amount_ident, &mut required_args);
+    let (setup_fn_optional_args, fn_optional_args) = contract_function_optional_args_tokens(
+        &receive_attributes.optional,
+        &amount_ident,
+        &mut required_args,
+    );
 
-    let mut out = if contains_attribute(&attrs, "low_level") {
+    let mut out = if receive_attributes.optional.low_level {
         required_args.push("state: &mut ContractState");
         quote! {
             #[export_name = #wasm_export_fn_name]
@@ -463,7 +738,7 @@ fn receive_worker(attr: TokenStream, item: TokenStream) -> syn::Result<TokenStre
     }
 
     // Embed schema if 'parameter' attribute is set
-    let parameter_option = get_attribute_value(attrs.iter(), "parameter")?;
+    let parameter_option = receive_attributes.optional.parameter.map(|a| a.value());
     out.extend(contract_function_schema_tokens(
         parameter_option,
         rust_export_fn_name,
@@ -480,25 +755,25 @@ fn receive_worker(attr: TokenStream, item: TokenStream) -> syn::Result<TokenStre
 ///
 /// It also mutates a vector of required arguments with the expected type
 /// signature of each.
-fn contract_function_optional_args_tokens<'a, I: Copy + IntoIterator<Item = &'a Meta>>(
-    attrs: I,
+fn contract_function_optional_args_tokens(
+    optional: &OptionalArguments,
     amount_ident: &syn::Ident,
     required_args: &mut Vec<&str>,
 ) -> (proc_macro2::TokenStream, Vec<proc_macro2::TokenStream>) {
     let mut setup_fn_args = proc_macro2::TokenStream::new();
     let mut fn_args = vec![];
-    if contains_attribute(attrs, "payable") {
+    if optional.payable {
         required_args.push("amount: Amount");
         fn_args.push(quote!(#amount_ident));
     } else {
         setup_fn_args.extend(quote! {
             if #amount_ident.micro_gtu != 0 {
-                return -1;
+                return concordium_std::Reject::from(concordium_std::NotPayableError).error_code.get();
             }
         });
     };
 
-    if contains_attribute(attrs, "enable_logger") {
+    if optional.enable_logger {
         required_args.push("logger: &mut impl HasLogger");
         let logger_ident = format_ident!("logger");
         setup_fn_args.extend(quote!(let mut #logger_ident = concordium_std::Logger::init();));
@@ -546,14 +821,14 @@ fn contract_function_schema_tokens(
 /// In addition to the attributes supported by `derive(Serial)`, this derivation
 /// macro supports the `ensure_ordered` attribute. If applied to a field the
 /// of type `BTreeMap` or `BTreeSet` deserialization will additionally ensure
-/// that there keys are in strictly increasing order. By default deserialization
+/// that the keys are in strictly increasing order. By default deserialization
 /// only ensures uniqueness.
 ///
 /// # Example
 /// ``` ignore
 /// #[derive(Deserial)]
 /// struct Foo {
-///     #[concordium(set_size_length = 1, ensure_ordered)]
+///     #[concordium(size_length = 1, ensure_ordered)]
 ///     bar: BTreeSet<u8>,
 /// }
 /// ```
@@ -567,8 +842,7 @@ pub fn deserial_derive(input: TokenStream) -> TokenStream {
 const CONCORDIUM_FIELD_ATTRIBUTE: &str = "concordium";
 
 /// A list of valid concordium field attributes
-const VALID_CONCORDIUM_FIELD_ATTRIBUTES: [&str; 5] =
-    ["size_length", "set_size_length", "map_size_length", "string_size_length", "ensure_ordered"];
+const VALID_CONCORDIUM_FIELD_ATTRIBUTES: [&str; 3] = ["size_length", "ensure_ordered", "rename"];
 
 fn get_concordium_field_attributes(attributes: &[syn::Attribute]) -> syn::Result<Vec<syn::Meta>> {
     attributes
@@ -630,94 +904,89 @@ fn find_field_attribute_value(
     }
 }
 
-fn find_length_attribute(
-    attributes: &[syn::Attribute],
-    target_attr: &str,
-) -> syn::Result<Option<u32>> {
-    let value = match find_field_attribute_value(attributes, target_attr)? {
+fn find_length_attribute(attributes: &[syn::Attribute]) -> syn::Result<Option<u32>> {
+    let value = match find_field_attribute_value(attributes, "size_length")? {
         Some(v) => v,
         None => return Ok(None),
     };
 
+    // Save the span to be used in errors.
+    let value_span = value.span();
+
     let value = match value {
         syn::Lit::Int(int) => int,
-        _ => {
-            return Err(syn::Error::new(value.span(), "Length attribute value must be an integer."))
-        }
+        _ => return Err(syn::Error::new(value_span, "Length attribute value must be an integer.")),
     };
     let value = match value.base10_parse() {
         Ok(v) => v,
         _ => {
             return Err(syn::Error::new(
-                value.span(),
+                value_span,
                 "Length attribute value must be a base 10 integer.",
             ))
         }
     };
     match value {
         1 | 2 | 4 | 8 => Ok(Some(value)),
-        _ => Err(syn::Error::new(value.span(), "Length info must be either 1, 2, 4, or 8.")),
+        _ => Err(syn::Error::new(value_span, "Length info must be either 1, 2, 4, or 8.")),
     }
 }
 
+/// Find a 'rename' attribute and return its value and span.
+/// Checks that the attribute is only defined once and that the value is a
+/// string.
+#[cfg(feature = "build-schema")]
+fn find_rename_attribute(attributes: &[syn::Attribute]) -> syn::Result<Option<(String, Span)>> {
+    let value = match find_field_attribute_value(attributes, "rename")? {
+        Some(v) => v,
+        None => return Ok(None),
+    };
+
+    match value {
+        syn::Lit::Str(value) => Ok(Some((value.value(), value.span()))),
+        _ => Err(syn::Error::new(value.span(), "Rename attribute value must be a string.")),
+    }
+}
+
+/// Check for name collisions by inserting the name in the HashMap.
+/// On collisions it returns a combined error pointing to the previous and new
+/// definition.
+#[cfg(feature = "build-schema")]
+fn check_for_name_collisions(
+    used_names: &mut HashMap<String, Span>,
+    new_name: &str,
+    new_span: Span,
+) -> syn::Result<()> {
+    if let Some(used_span) = used_names.insert(String::from(new_name), new_span) {
+        let error_msg = format!("the name `{}` is defined multiple times", new_name);
+        let mut error_at_first_def = syn::Error::new(used_span, &error_msg);
+        let error_at_second_def = syn::Error::new(new_span, &error_msg);
+
+        // Combine the errors to show both at once
+        error_at_first_def.combine(error_at_second_def);
+
+        return Err(error_at_first_def);
+    }
+    Ok(())
+}
+
 fn impl_deserial_field(
     f: &syn::Field,
     ident: &syn::Ident,
     source: &syn::Ident,
 ) -> syn::Result<proc_macro2::TokenStream> {
     let concordium_attributes = get_concordium_field_attributes(&f.attrs)?;
-    if let Some(l) = find_length_attribute(&f.attrs, "size_length")? {
-        let size = format_ident!("u{}", 8 * l);
+    let ensure_ordered = contains_attribute(&concordium_attributes, "ensure_ordered");
+    let size_length = find_length_attribute(&f.attrs)?;
+    let has_ctx = ensure_ordered || size_length.is_some();
+    let ty = &f.ty;
+    if has_ctx {
+        // Default size length is u32, i.e. 4 bytes.
+        let l = format_ident!("U{}", 8 * size_length.unwrap_or(4));
         Ok(quote! {
-            let #ident = {
-                let len = #size::deserial(#source)?;
-                deserial_vector_no_length(#source, len as usize)?
-            };
-        })
-    } else if let Some(l) = find_length_attribute(&f.attrs, "map_size_length")? {
-        let size = format_ident!("u{}", 8 * l);
-        if contains_attribute(&concordium_attributes, "ensure_ordered") {
-            Ok(quote! {
-                let #ident = {
-                    let len = #size::deserial(#source)?;
-                    deserial_map_no_length(#source, len as usize)?
-                };
-            })
-        } else {
-            Ok(quote! {
-                let #ident = {
-                    let len = #size::deserial(#source)?;
-                    deserial_map_no_length_no_order_check(#source, len as usize)?
-                };
-            })
-        }
-    } else if let Some(l) = find_length_attribute(&f.attrs, "set_size_length")? {
-        let size = format_ident!("u{}", 8 * l);
-        if contains_attribute(&concordium_attributes, "ensure_ordered") {
-            Ok(quote! {
-                let #ident = {
-                    let len = #size::deserial(#source)?;
-                    deserial_set_no_length(#source, len as usize)?
-                };
-            })
-        } else {
-            Ok(quote! {
-                let #ident = {
-                    let len = #size::deserial(#source)?;
-                    deserial_set_no_length_no_order_check(#source, len as usize)?
-                };
-            })
-        }
-    } else if let Some(l) = find_length_attribute(&f.attrs, "string_size_length")? {
-        let size = format_ident!("u{}", 8 * l);
-        Ok(quote! {
-            let #ident = {
-                let len = #size::deserial(#source)?;
-                deserial_string(#source, len as usize)?
-            };
+            let #ident = <#ty as concordium_std::DeserialCtx>::deserial_ctx(concordium_std::schema::SizeLength::#l, #ensure_ordered, #source)?;
         })
     } else {
-        let ty = &f.ty;
         Ok(quote! {
             let #ident = <#ty as Deserial>::deserial(#source)?;
         })
@@ -844,11 +1113,10 @@ fn impl_deserial(ast: &syn::DeriveInput) -> syn::Result<TokenStream> {
 /// Serial trait.
 ///
 ///
-/// Collections (Vec, BTreeMap, BTreeSet) are by default serialized by
-/// prepending the number of elements as 4 bytes little-endian. If this is too
-/// much or too little, fields of the above types can be annotated with
-/// `size_length` for Vec, `map_size_length` for `BTreeMap` and
-/// `set_size_length` for `BTreeSet`.
+/// Collections (Vec, BTreeMap, BTreeSet) and strings (String, str) are by
+/// default serialized by prepending the number of elements as 4 bytes
+/// little-endian. If this is too much or too little, fields of the above types
+/// can be annotated with `size_length`.
 ///
 /// The value of this field is the number of bytes that will be used for
 /// encoding the number of elements. Supported values are 1, 2, 4, 8.
@@ -868,7 +1136,7 @@ fn impl_deserial(ast: &syn::DeriveInput) -> syn::Result<TokenStream> {
 /// ```ignore
 /// #[derive(Serial)]
 /// struct Foo {
-///     #[concordium(set_size_length = 1)]
+///     #[concordium(size_length = 1)]
 ///     bar: BTreeSet<u8>,
 /// }
 /// ```
@@ -883,34 +1151,12 @@ fn impl_serial_field(
     ident: &proc_macro2::TokenStream,
     out: &syn::Ident,
 ) -> syn::Result<proc_macro2::TokenStream> {
-    if let Some(l) = find_length_attribute(&field.attrs, "size_length")? {
-        let id = format_ident!("u{}", 8 * l);
-        Ok(quote! {
-            let len: #id = #ident.len() as #id;
-            len.serial(#out)?;
-            serial_vector_no_length(&#ident, #out)?;
-        })
-    } else if let Some(l) = find_length_attribute(&field.attrs, "map_size_length")? {
-        let id = format_ident!("u{}", 8 * l);
-        Ok(quote! {
-            let len: #id = #ident.len() as #id;
-            len.serial(#out)?;
-            serial_map_no_length(&#ident, #out)?;
-        })
-    } else if let Some(l) = find_length_attribute(&field.attrs, "set_size_length")? {
-        let id = format_ident!("u{}", 8 * l);
-        Ok(quote! {
-            let len: #id = #ident.len() as #id;
-            len.serial(#out)?;
-            serial_set_no_length(&#ident, #out)?;
-        })
-    } else if let Some(l) = find_length_attribute(&field.attrs, "string_size_length")? {
-        let id = format_ident!("u{}", 8 * l);
-        Ok(quote! {
-            let len: #id = #ident.len() as #id;
-            len.serial(#out)?;
-            serial_string(#ident.as_str(), #out)?;
-        })
+    if let Some(size_length) = find_length_attribute(&field.attrs)? {
+        let l = format_ident!("U{}", 8 * size_length);
+        Ok(quote!({
+            use concordium_std::SerialCtx;
+            #ident.serial_ctx(concordium_std::schema::SizeLength::#l, #out)?;
+        }))
     } else {
         Ok(quote! {
             #ident.serial(#out)?;
@@ -1086,15 +1332,9 @@ fn contract_state_worker(attr: TokenStream, item: TokenStream) -> syn::Result<To
 
     let attrs = Punctuated::<Meta, Token![,]>::parse_terminated.parse(attr)?;
 
-    let contract_name = get_attribute_value(attrs.iter(), "contract")?.ok_or_else(|| {
-        syn::Error::new(
-            attrs.span(),
-            "A name of the contract must be provided, using the 'contract' attribute.\n\nFor \
-             example #[contract_state(contract = \"my-contract\")].",
-        )
-    })?;
-
-    let wasm_schema_name = format!("concordium_schema_state_{}", contract_name);
+    let contract_state_attributes = parse_contract_state_attributes(&attrs)?;
+    let contract_name = contract_state_attributes.contract;
+    let wasm_schema_name = format!("concordium_schema_state_{}", contract_name.value());
     let rust_schema_name = format_ident!("concordium_schema_state_{}", data_ident);
 
     let generate_schema_tokens = quote! {
@@ -1118,10 +1358,7 @@ fn contract_state_worker(_attr: TokenStream, item: TokenStream) -> syn::Result<T
 /// Derive the `SchemaType` trait for a type.
 /// If the feature `build-schema` is not enabled this is a no-op, i.e., it does
 /// not produce any code.
-#[proc_macro_derive(
-    SchemaType,
-    attributes(size_length, map_size_length, set_size_length, string_size_length)
-)]
+#[proc_macro_derive(SchemaType, attributes(size_length))]
 pub fn schema_type_derive(input: TokenStream) -> TokenStream {
     unwrap_or_report(schema_type_derive_worker(input))
 }
@@ -1132,6 +1369,8 @@ fn schema_type_derive_worker(input: TokenStream) -> syn::Result<TokenStream> {
 
     let data_name = &ast.ident;
 
+    let (impl_generics, ty_generics, where_clauses) = ast.generics.split_for_impl();
+
     let body = match ast.data {
         syn::Data::Struct(ref data) => {
             let fields_tokens = schema_type_fields(&data.fields)?;
@@ -1140,11 +1379,23 @@ fn schema_type_derive_worker(input: TokenStream) -> syn::Result<TokenStream> {
             }
         }
         syn::Data::Enum(ref data) => {
+            let mut used_variant_names = HashMap::new();
             let variant_tokens: Vec<_> = data
                 .variants
                 .iter()
                 .map(|variant| {
-                    let variant_name = &variant.ident.to_string();
+                    // Handle the 'rename' attribute.
+                    let (variant_name, variant_span) = match find_rename_attribute(&variant.attrs)?
+                    {
+                        Some(name_and_span) => name_and_span,
+                        None => (variant.ident.to_string(), variant.ident.span()),
+                    };
+                    check_for_name_collisions(
+                        &mut used_variant_names,
+                        &variant_name,
+                        variant_span,
+                    )?;
+
                     let fields_tokens = schema_type_fields(&variant.fields)?;
                     Ok(quote! {
                         (concordium_std::String::from(#variant_name), #fields_tokens)
@@ -1160,7 +1411,7 @@ fn schema_type_derive_worker(input: TokenStream) -> syn::Result<TokenStream> {
 
     let out = quote! {
         #[automatically_derived]
-        impl concordium_std::schema::SchemaType for #data_name {
+        impl #impl_generics concordium_std::schema::SchemaType for #data_name #ty_generics #where_clauses {
             fn get_type() -> concordium_std::schema::Type {
                 #body
             }
@@ -1174,27 +1425,10 @@ fn schema_type_derive_worker(_input: TokenStream) -> syn::Result<TokenStream> {
     Ok(TokenStream::new())
 }
 
-#[cfg(feature = "build-schema")]
-fn or_else_joined<A>(
-    a: syn::Result<Option<A>>,
-    b: impl FnOnce() -> syn::Result<Option<A>>,
-) -> syn::Result<Option<A>> {
-    match a {
-        Ok(None) => b(),
-        _ => a,
-    }
-}
-
 #[cfg(feature = "build-schema")]
 fn schema_type_field_type(field: &syn::Field) -> syn::Result<proc_macro2::TokenStream> {
     let field_type = &field.ty;
-    if let Some(l) = or_else_joined(find_length_attribute(&field.attrs, "size_length"), || {
-        or_else_joined(find_length_attribute(&field.attrs, "map_size_length"), || {
-            or_else_joined(find_length_attribute(&field.attrs, "set_size_length"), || {
-                find_length_attribute(&field.attrs, "string_size_length")
-            })
-        })
-    })? {
+    if let Some(l) = find_length_attribute(&field.attrs)? {
         let size = format_ident!("U{}", 8 * l);
         Ok(quote! {
             <#field_type as concordium_std::schema::SchemaType>::get_type().set_size_length(concordium_std::schema::SizeLength::#size)
@@ -1210,10 +1444,17 @@ fn schema_type_field_type(field: &syn::Field) -> syn::Result<proc_macro2::TokenS
 fn schema_type_fields(fields: &syn::Fields) -> syn::Result<proc_macro2::TokenStream> {
     match fields {
         syn::Fields::Named(_) => {
+            let mut used_field_names = HashMap::new();
             let fields_tokens: Vec<_> = fields
                 .iter()
                 .map(|field| {
-                    let field_name = field.ident.clone().unwrap().to_string(); // safe since named fields
+                    // Handle the 'rename' attribute.
+                    let (field_name, field_span) = match find_rename_attribute(&field.attrs)? {
+                        Some(name_and_span) => name_and_span,
+                        None => (field.ident.clone().unwrap().to_string(), field.ident.span()), // safe since named fields.
+                    };
+                    check_for_name_collisions(&mut used_field_names, &field_name, field_span)?;
+
                     let field_schema_type = schema_type_field_type(&field)?;
                     Ok(quote! {
                         (concordium_std::String::from(#field_name), #field_schema_type)
